import { createContext } from 'react';
/**
`AppContext` is a React context that exposes a method to manually exit the app (unmount).
*/
export const AppContext = createContext({
    exit() { },
    rerender() { },
    options: {},
    setOptions() { },
    dumpCurrentFrame() { },
    startRecording() { },
    stopRecording() { },
});
AppContext.displayName = 'AppContext';
export default AppContext;
//# sourceMappingURL=AppContext.js.map